# -*- coding: utf-8 -*-
# Author: Douglas Creager <dcreager@dcreager.net>
# Modified: Andrew Mark
#
# Updates the package version number to reflect the git tag of the current commit.
#
# There are two normal use cases:
#  python setup.py sdist: Building a source distribution. This is expected to
#         take place in a folder tracked by git. The return value of "git --tags"
#         is taken as the version number, and this is stored in version_file
#         as part of the distribution.
#  python setup.py install: Here the version number is taken from the value
#         stored in version_file
#
# The file version_file should not be tracked. Doing so will lead to version
# recursion, since it is updated only after the latest version has been
# committed and tagged.
#
# To use this script, simply import it your setup.py file, and use the
# results of get_version() as your package version, and add the locally defined
# command 'sdist' as a cmdclass:
#
# from git_version import get_version, sdist
#
# setup(
#     version=get_git_version(),
#     cmdclass={"sdist": sdist },
#     .
#     .
# )

__all__ = ("get_version",)

from distutils.command.sdist import sdist as _sdist
from subprocess import Popen, PIPE
import re

version_file = 'gdsCAD/_version.py'

def git_version():
    """Return the git tag of the current commit"""
    try:
        p = Popen(['git', 'describe', '--tags'],
                  stdout=PIPE, stderr=PIPE)
        p.stderr.close()
        line = p.stdout.readlines()[0]
        return line.strip()

    except:
        return None

def file_version():
    """Return the version number stored in _version.py"""
    try:
        f = open(version_file)
    except EnvironmentError:
        return None
    for line in f.readlines():
        mo = re.match("__version__ = '([^']+)'", line)
        if mo:
            ver = mo.group(1)
            return ver
    return None


VERSION_PY = """
# This version number is untracked and may not be up to date. 
# It is updated during source builds.

__version__ = '%s'
"""

def update_f_version(new_ver):
    """Change the version number stored in _version.py"""
    with open(version_file, "w") as f:
        f.write(VERSION_PY % new_ver)

class sdist(_sdist):
    """New sdist command that updates the version number stored in _version.py"""
    def run(self):
        version = get_version()        
        update_f_version(version)
        self.distribution.metadata.version = version
        return _sdist.run(self)

def get_version():
    # Read in the version that's currently in RELEASE-VERSION.
    f_version = file_version()

    # First try to get the current version using “git describe”.
    version = git_version()

    # If that doesn't work, fall back on the value that's in
    # RELEASE-VERSION.
    if version is None:
        version = f_version

    # If we still don't have anything, that's an error.
    if version is None:
        raise ValueError("Cannot find the version number!")

    # Finally, return the current version.
    return version


if __name__ == "__main__":
    print get_version()



# -*- coding: utf-8 -*-
"""
Classes to define simple shapes.

Filled Objects
--------------
:class:`Rectangle`
    A filled rectangle
:class:`Disk`
    A filled circle
:class:`RegPolygon`
    A filled regular polygon
:class:`Label`
    Printing text

Unfilled Objects
----------------
:class:`Box`
    An unfilled rectangle
:class:`RegPolyline`
    An ufilled regular polygon
:class:`Circle`
    An unfilled circle


.. note::
    Copyright 2009-2012 Lucas Heitzmann Gabrielli
    
    Copyright 2013 Andrew G. Mark

    gdsCAD (based on gdspy) is released under the terms of the GNU GPL
    
"""

import os
import numpy as np

import core


class Rectangle(core.Boundary):
    """
    Filled rectangular geometric object.

    :param point1: Coordinates of a corner of the rectangle.
    :param point2: Coordinates of the corner of the rectangle opposite to ``point1``.
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).

    Examples::

        rectangle = shapes.Rectangle((0, 0), (10, 20))
        myCell.add(rectangle)
    """
    
    def __init__(self, point1, point2, layer=None, datatype=None):
        
        points = np.array([[point1[0], point1[1]], [point1[0], point2[1]], [point2[0], point2[1]], [point2[0], point1[1]]])        
        core.Boundary.__init__(self, points,  layer, datatype)        

    def __str__(self):
        return "Rectangle (({0[0]}, {0[1]}) to ({1[0]}, {1[1]}), layer {2}, datatype {3})".format(self.points[0], self.points[2], self.layer, self.datatype)

    def __repr__(self):
        return "Rectangle({2}, ({0[0]}, {0[1]}), ({1[0]}, {1[1]}), {3})".format(self.points[0], self.points[2], self.layer, self.datatype)


class Box(core.Path):
    """
    Unfilled rectangular geometric object.

    :param point1: Coordinates of a corner of the rectangle.
    :param point2: Coordinates of the corner of the rectangle opposite to ``point1``.
    :param width: The width of the line
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).

    Examples::
        
        box = shapes.Box((0, 0), (10, 20), 0.5)
        myCell.add(box)
    """
    
    def __init__(self, point1, point2, width, layer=None, datatype=None):
        
        points = np.array([[point1[0], point1[1]], [point1[0], point2[1]], [point2[0], point2[1]], [point2[0], point1[1]], [point1[0], point1[1]]])        
        core.Path.__init__(self, points, width, layer, datatype)        

    def __str__(self):
        return "Box (({0[0]}, {0[1]}) to ({1[0]}, {1[1]}), layer {2}, datatype {3})".format(self.points[0], self.points[2], self.layer, self.datatype)

    def __repr__(self):
        return "Box ({2}, ({0[0]}, {0[1]}), ({1[0]}, {1[1]}), {3})".format(self.points[0], self.points[2], self.layer, self.datatype)


class Disk(core.Boundary):
    """
    A filled circle, or section of a circle

    :param center: Coordinates of the disk's center.
    :param radius: The radius of the disk
    :param inner_radius: The inner radius of the disk. If absent creates a solid disk.
    :param initial_angle: The starting angle of the sweep
    :param final_angle: The final angle of the sweep
    :param number_of_points: The number of line segments that the disk will be composed of
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).

    Example::
        
        disk=shapes.Disk((-5,-5), 5)
        disk.show()    
    """


    def __init__(self, center, radius, inner_radius=0, initial_angle=0, final_angle=0, number_of_points=199, layer=None, datatype=None):

        self.center = center
        self.radius = radius

        if final_angle == initial_angle:
            final_angle += 360.0
            
        angles = np.linspace(initial_angle, final_angle, number_of_points).T * np.pi/180.

        points=np.vstack((np.cos(angles), np.sin(angles))).T * radius + np.array(center)

        if inner_radius != 0:
            points2 = np.vstack((np.cos(angles), np.sin(angles))).T * inner_radius + np.array(center)
            points=np.vstack((points, points2[::-1]))
        
        core.Boundary.__init__(self, points, layer, datatype)
        
    def __str__(self):
        return "Disk Boundary (center={}, radius={}, layer={}, datatype={})".format(self.center, self.radius, self.layer, self.datatype)
    
    
class Circle(core.Path):
    """
    An unfilled circular path or section or arc.

    :param center: Coordinates of the disk's center.
    :param radius: The radius of the disk.
    :param width: The width of the line.
    :param initial_angle: The starting angle of the sweep.
    :param final_angle: The final angle of the sweep.
    :param number_of_points: The number of line segments that the disk will be composed of.
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).
    
    Example::
        
        circ=shapes.Circle((10,10), 10, 0.5)
        circ.show()
    """

    def __init__(self, center, radius, width, initial_angle=0, final_angle=0, number_of_points=199, layer=None, datatype=None):

        self.center = center
        self.radius = radius

        if final_angle == initial_angle:
            final_angle += 360.0
            
        angles = np.linspace(initial_angle, final_angle, number_of_points) * np.pi/180.

        points=np.vstack((np.cos(angles), np.sin(angles))).T * radius + np.array(center)

        core.Path.__init__(self, points, width, layer, datatype)


    def __str__(self):
        return "Circle Path ({} points, width {}, layer {}, datatype {})".format(len(self.points), self.width, self.layer, self.datatype)

class RegPolygon(core.Boundary):
    """
    An unfilled regular polgyon.

    :param center: Coordinates of the disk's center.
    :param length: The length of an edge.
    :param N: The number of sides
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).
    
    Example::
        
        pent = shapes.RegPolygon((10,10), 10, 5)
        pent.show()
    """

    def __init__(self, center, length, N, layer=None, datatype=None):

        self.center = center
        self.length = length
        self.N = N

        angles = np.linspace(0, 360, N, endpoint=False) * np.pi/180.

        alpha = angles[1]
        radius = length / np.sin(alpha/2) /2.
        points=np.vstack((np.cos(angles), np.sin(angles))).T * radius + np.array(center)

        core.Boundary.__init__(self, points, layer, datatype)


    def __str__(self):
        return "RegPolygon Boundary ({} points, width {}, layer {}, datatype {})".format(len(self.points), self.width, self.layer, self.datatype)


class RegPolyline(core.Path):
    """
    An unfilled regular polgyon.

    :param center: Coordinates of the disk's center.
    :param length: The length of an edge.
    :param N: The number of sides
    :param width: The width of the line.
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).
    
    Example::
        
        hex=shapes.RegPolylone((10,10), 10, 6, 0.5)
        hex.show()
    """

    def __init__(self, center, length, N, width, layer=None, datatype=None):

        self.center = center
        self.length = length
        self.N = N

        angles = np.linspace(0, 360, N+1, endpoint=True) * np.pi/180.

        alpha = angles[1]
        radius = length / np.sin(alpha/2) /2.
        points=np.vstack((np.cos(angles), np.sin(angles))).T * radius + np.array(center)

        core.Path.__init__(self, points, width, layer, datatype)


    def __str__(self):
        return "RegPolyine Path ({} points, width {}, layer {}, datatype {})".format(len(self.points), self.width, self.layer, self.datatype)


class Label(core.Elements):
    """
    Printing text string object.
    
    Each letter is formed by a series of polygons collected together as an
    Elements list.

    :param text: The text to be converted in geometric objects.
    :param size: Base size of each character.
    :param position: Text position (lower left corner).
    :param horizontal: If ``True``, the text is written from left to right;
      if ``False``, from top to bottom.
    :param angle: The angle of rotation of the text.
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).

    Examples::
        
        text = shapes.Label('Sample text', 20, (-10, -100))
        text.show()
        myCell.add(text)
    """
    from font import _font

    def __init__(self, text, size, position=(0, 0), horizontal=True, angle=0, layer=None, datatype=None) :

        self.text=text
        self.position=position

        polygons = []
        posX = 0
        posY = 0
        text_multiplier = size / 9.0

        for jj in range(len(text)):
            if text[jj] == '\n':
                if horizontal:
                    posY -= 11
                    posX = 0
                else:
                    posX += 8
                    posY = 0
            elif text[jj] == '\t':
                if horizontal:
                    posX = posX + 32 - (posX + 8) % 32
                else:
                    posY = posY - 11 - (posY - 22) % 44
            else:
                if Label._font.has_key(text[jj]):
                    for p in Label._font[text[jj]]:
                        polygon = p[:]
                        for ii in range(len(polygon)):
                            xp = text_multiplier * (posX + polygon[ii][0])
                            yp = text_multiplier * (posY + polygon[ii][1])
                            polygon[ii] = (xp, yp)
                        polygons.append(np.array(polygon))
                if horizontal:
                    posX += 8
                else:
                    posY -= 11
        core.Elements.__init__(self, polygons, layer, datatype)
        self.rotate(angle)
        self.translate(position)

    def __str__(self):
        return "Text -\"{}\" layer={}".format(self.text, self.layer)


class LineLabel(core.Elements):
    """
    Printing text string object as Line font.

    Each letter is formed by a series of lines collected together as an
    Elements list. The lines are Hershey vector fonts. The font is basically
    a collection of symbols, containing all kinds of font styles and symbols.

    The font itself is not a monotype and no guarantee is given on the height
    of the characters. But for normal characters, the height should always be
    within the specified size.

    Since the Hershey nearly as old as ASCII, it is unfortunately not ordered
    in these codes. The font itself also contains many styles, ranging from
    Cyrillic over Greek to Roman letters.

    For convenience, lookup tables from ASCII to symbol number are provided
    for several font styles. But note, that only a fraction of all available
    symbols are mapped to ASCII characters in these tables. You can always add
    a specific symbol by yourself, once you know its Hershey code.

    +-------------+--------------------------+
    | Table name  | Description              |
    +=============+==========================+
    | gothgbt     | Gothic English Triplex   |
    +-------------+--------------------------+
    | gothgrt     | Gothic German Triplex    |
    +-------------+--------------------------+
    | gothitt     | Gothic Italian Triplex   |
    +-------------+--------------------------+
    | greekc      | Greek Complex            |
    +-------------+--------------------------+
    | greekcs     | Greek Complex Small      |
    +-------------+--------------------------+
    | greekp      | Greek Plain              |
    +-------------+--------------------------+
    | greeks      | Greek Simplex            |
    +-------------+--------------------------+
    | cyrilc      | Cyrillic complex         |
    +-------------+--------------------------+
    | italicc     | Italic Complex           |
    +-------------+--------------------------+
    | italiccs    | Italic Complex Small     |
    +-------------+--------------------------+
    | italict     | Italic Triplex           |
    +-------------+--------------------------+
    | scriptc     | Script Complex           |
    +-------------+--------------------------+
    | scripts     | Script Simplex           |
    +-------------+--------------------------+
    | romanc      | Roman Complex            |
    +-------------+--------------------------+
    | romancs     | Roman Complex Small      |
    +-------------+--------------------------+
    | romand      | Roman Duplex             |
    +-------------+--------------------------+
    | romanp      | Roman Plain              |
    +-------------+--------------------------+
    | romans      | Roman Simplex            |
    +-------------+--------------------------+
    | romant      | Roman Triplex            |
    +-------------+--------------------------+

    :param text: The text to be converted in geometric objects.
    :param size: Base size of each character.
    :param position: Text position (lower left corner).
    :param style: The default name of ASCII lookup table.
        Defaults to (romans) Roman Simplex.
    :param horizontal: If ``True``, the text is written from left to right;
      if ``False``, from top to bottom.
    :param line_width: Line width of the text.
        Defaults to one 40th of the text size.
    :param layer: The GDSII layer number for this element.
        Defaults to layer of 1st object, or core.default_layer.
    :param datatype: The GDSII datatype for this element (between 0 and 255).

    Examples::

        text = shapes.LineLabel(20, (-10, -100))
        text.add_text('Sample text', 'romand')
        text.show()
        myCell.add(text)
    """

    _DEFAULT_CHAR_HEIGHT = 40.

    _hershey_table = dict()
    _hershey_ascii_lookup_table = dict()

    def __init__(self, text, size, position=(0, 0), style='romans', horizontal=True,
                 line_width=None, layer=None, datatype=None):
        if not len(self._hershey_table):
            self._load_hersey_table()

        self._scale = size / self._DEFAULT_CHAR_HEIGHT
        self._line_width = line_width if line_width else size/40.
        self._layer = layer
        self._datatype = datatype
        self._origin = position
        self._style = style
        self._horizontal = horizontal
        self._symbols = list()
        self._symbol_pos = list(position)

        core.Elements.__init__(self)

        self.add_text(text, style)

    def _load_hersey_table(self):
        """
        Load the hersey table.

        This modifies a static class variable thus avoiding to
        read the Hersey table multiple times.
        """
        self._hershey_table.clear()

        path, _ = os.path.split(__file__)
        fname = os.path.join(path, 'resources', 'hershey', 'hershey')

        # Read lines and correct newlines
        lines = []
        append_next_line = False
        vertices_count = 0
        target_vertices_count = 0
        for line in open(fname, 'r').readlines():
            line = line.rstrip('\n')

            # Skip empty lines
            if not len(line):
                continue

            if vertices_count == target_vertices_count:
                # New entry begins here
                target_vertices_count = int(line[5:8]) * 2
                vertices_count = len(line[8:])
                lines.append(line)
            else:
                assert vertices_count < target_vertices_count, 'Got more vertices then specified'
                vertices_count += len(line)
                lines[-1] += line

        # Parse the lines we have just read
        for line in lines:
            # Skip empty lines
            if not len(line):
                continue

            char_id = int(line[:5])
            n_vertices = int(line[5:8])

            left, right = [ord(c) - ord('R') for c in line[8:10]]
            vertices = line[10:]
            assert len(vertices) % 2 == 0, 'Number coordinates needs to be even since it comes in pairs'

            # Convert ascii coordinates to true paths
            strokes = list()
            current_stroke = list()
            for coordinate in zip(vertices[::2], vertices[1::2]):
                if coordinate[0] == ' ' and coordinate[1] == 'R':
                    strokes.append(np.array(current_stroke, dtype=np.int))
                    current_stroke = list()
                    continue

                tmp_coords = [ord(coordinate[i]) - ord('R') for i in (0, 1)]
                tmp_coords[1] = -tmp_coords[1]
                current_stroke.append(tmp_coords)

            strokes.append(np.array(current_stroke, dtype=np.int))

            self._hershey_table[char_id] = {'strokes': strokes, 'left_pos': left, 'right_pos': right}

    def _load_hersey_ascii_lookup_table(self, table_name):
        """
        Load an ASCII lookup table for the given font name.

        This modifies a static class variable thus avoiding to read the
        lookup table multiple times.

        :param table_name: The table name of the lookup table.
        """
        path, _ = os.path.split(__file__)
        fname = os.path.join(path, 'resources', 'hershey', '%s.hmp' % table_name)

        to_ascii = dict()
        from_ascii = dict()

        current_code = 32
        for line in open(fname, 'r').readlines():
            for definition in line.rstrip().split():
                # Either one number or a range of number is specified
                if not '-' in definition:
                    to_ascii[int(definition)] = current_code
                    from_ascii[current_code] = int(definition)
                    current_code += 1
                else:
                    start, stop = [int(x) for x in definition.split('-')]
                    for i in range(start, stop+1):
                        to_ascii[i] = current_code
                        from_ascii[current_code] = i
                        current_code += 1

        self._hershey_ascii_lookup_table[table_name] = {'to_ascii': to_ascii, 'from_ascii': from_ascii}

    def _add_single_symbol(self, symbol):
        """
        Internal function which adds one single symbol

        :param symbol: Symbol code
        """
        # Keep track of symbols to know what we have painted here
        self._symbols.append(symbol)

        assert symbol in self._hershey_table, 'This symbol is not in the Hershey table'
        symbol = self._hershey_table[symbol]

        for stroke in symbol['strokes']:
            if not len(stroke):
                continue
            scaled_stroke = (stroke + [0, self._DEFAULT_CHAR_HEIGHT/2]) * self._scale
            stroke_path = core.Path(scaled_stroke, self._line_width, self.layer,
                                    pathtype=2, datatype=self.datatype)
            stroke_path.translate([self._symbol_pos[0] - symbol['left_pos'] * self._scale,
                                   self._symbol_pos[1]])
            self.add(stroke_path)

        if self._horizontal:
            self._symbol_pos[0] += (symbol['right_pos'] - symbol['left_pos']) * self._scale
        else:
            self._symbol_pos[0] = self._origin[0]
            self._symbol_pos[1] -= self._DEFAULT_CHAR_HEIGHT * self._scale

    def add_symbol(self, symbol):
        """
        Add one or more symbols to the label.

        :param symbol: The integer symbol code as specified in the
            Hershey sign table.
        """
        if hasattr(symbol, '__iter__'):
            for x in symbol:
                self._add_single_symbol(x)
        else:
            self._add_single_symbol(symbol)

    def add_text(self, text, style=None):
        """
        Add text to the label.

        The corresponding symbol codes are automatically looked in
        the specified lookup table.

        :param text: The text to be converted in geometric objects.
        :param style: The name of the lookup table.
            Defaults to the style passed during the creation of this
            object.
        """

        style = style if style else self._style

        if style not in self._hershey_ascii_lookup_table:
            self._load_hersey_ascii_lookup_table(style)

        lookup_table = self._hershey_ascii_lookup_table[style]['from_ascii']
        for char in str(text):
            if char == '\n':
                self._symbol_pos[0] = self._origin[0]
                self._symbol_pos[1] -= self._DEFAULT_CHAR_HEIGHT * self._scale
                continue

            ascii_value = ord(char)
            if ascii_value in lookup_table:
                self._add_single_symbol(lookup_table[ascii_value])

    def __str__(self):
        text = ''
        for symbol in self._symbols:
            ascii_found = False
            for table in self._hershey_ascii_lookup_table.values():
                if symbol in table['to_ascii']:
                    ascii_found = True
                    text += chr(table['to_ascii'][symbol])
                    break

            if not ascii_found:
                text += '{%i}' % symbol

        return "VectorText -\"{}\" layer={}".format(text, self.layer)



# -*- coding: utf-8 -*-
"""
Utility functions for geometric transformations and layer manipulation.


.. note::
    Copyright 2009-2012 Lucas Heitzmann Gabrielli
    
    Copyright 2013 Andrew G. Mark

    gdsCAD (based on gdspy) is released under the terms of the GNU GPL
    
"""


import numpy as np
from core import (Cell, CellReference, CellArray,
                  ElementBase, Elements, ReferenceBase)

def translate(obj, displacement):
    """
    Translate an object, 2D vector, or a sequence of 2D vectors by the given vector

    :param obj: an object, a 2D vector, or a sequence of 2D vectors
    :param displacment: the vector by which to move ``obj``
    :returns: A moved copy of the ``obj``
    
    """

    if isinstance(obj, (ElementBase, Elements)):
        obj=obj.copy()
        obj.translate(displacement)
        return obj

    return np.array(obj)+np.array(displacement)


def rotate(obj, theta, center=(0,0)):
    """
    Rotate an object by given angle

    :param obj: an object, a 2D vector, or a sequence of 2D vectors
    :param theta: angle by which to rotate points in deg
    :param center: optional, pt about which to perform the rotation 
    :return: A rotated copy of ``obj``
    """
    if isinstance(obj, (ElementBase, Elements)):
        obj=obj.copy()
        obj.rotate(theta, center)
        return obj

    pts=np.array(obj)
    ang = theta * np.pi/180
    m=np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])

    if isinstance(center, str) and center.lower()=='com':
        center=pts.mean(0)
    else:    
        center=np.array(center)

    return m.dot((np.array(pts)-center).T).T+center

def reflect(obj, axis, origin=(0,0), reverse_seq=True):
    """
    Reflect an object in the x or y axis

    :param obj: a geometric object, 2D vector, or a sequence of 2D vectors.
    :param axis: string 'x' or 'y' indcating which axis in which to make the refln.
    :param origin: optional, pt about which to perform the rotation.
    :param reverse_seq: if ``True`` reverses the order of a sequence.
    :returns: a reflected copy of ``obj``

    Sequences of points are reversed to maintain the same sense as the
    original sequence.    

    """
    if isinstance(obj, (ElementBase, Elements)):
        obj=obj.copy()
        obj.reflect(axis, origin)
        return obj


    if axis=='x':
        return scale(obj, [1,-1], origin, reverse_seq)
    elif axis=='y':
        return scale(obj, [-1,1], origin, reverse_seq)
    else:
        raise ValueError('Unknown axis %s'%str(axis))


def scale(obj, k, origin=(0,0), reverse_seq=True):
    """
    Scale the pt or sequence of pts by the factor k

    :param obj: a geometric object, 2D vector, or a sequence of 2D vectors.
    :param k: factor by which to scale ``obj``
    :param origin: pt which remains invariant under scale
    :param reverse_seq: if ``True`` reverses the order of a sequence.
    :returns: a scaled copy of ``obj``
    
    The factor k can be a scalar or 2D vector allowing non-uniform scaling
    Optional origin can be a 2D vector or 'COM' indicating that scaling should
    be made about the pts centre of mass. The center of mass is calculated
    only for the points, not by the area of the polygon they define.
    
    Sequences of points are reversed if necessary to maintain the
    same sense as the original sequence.    
    
    """
    if isinstance(obj, (ElementBase, Elements)):
        obj=obj.copy()
        obj.scale(k, origin)
        return obj

    pts=np.array(obj)
    if isinstance(origin, str) and origin.lower()=='com':
        origin=pts.mean(0)
    else:    
        origin=np.array(origin)
        
    k=np.array(k)
    
    if reverse_seq and ((k.prod()>=0) or k.shape==(2,)): #even parity or single point
        return (pts-origin)*k+origin
    else:
        return ((pts-origin)*k+origin)[::-1]


def split_layers(cell, old_layers):
    """
    Split the artwork in a cell between two copies according by layer.
          
    :param cells: The :class:`Cell` to split
    :param old_layers: A list of layers whose artwork will be moved to the second cell
    :returns: A tuple of two cells

    TODO: use same labelling scheme found in GdsImport
            
    :func:`split_layers` provides a convenient method of separating layers in 
    a ``Cell`` into different Cells. It creates two copies of ``cell`` based on
    the list of layers in ``old_layers``. Any artwork found on the layers of
    ``old_layers`` is moved to the first cell, and any artwork not found on the
    layers of ``old_layers`` is moved to the second layer. The existing heirarchy
    is maintained, however any empty ``Cells`` or references are removed.

    """
    
    subA=cell.copy()
    subB=cell.copy()

    #identify all art in subA that should be removed        
    blacklist=set()
    deps=subA.get_dependencies(True)
    print 'DEPENDENCY LIST HAS LENGTH: ',len(deps)
    for e in deps:
        if not isinstance(e, (Cell, CellReference, CellArray)):
            if e.layer in old_layers:
                blacklist.add(e)

    #remove references to removed art
    for c in [subA]+deps:
        if isinstance(c, Cell):
            c._objects=[e for e in c.objects if e not in blacklist]
    
    #clean heirarchy
    subA.prune()
            
    #identify all art in subB that should be removed        
    blacklist=set()
    deps=subB.get_dependencies(True)
    for e in deps:
        if not isinstance(e, (Cell, CellReference, CellArray)):
            if e.layer not in old_layers:
                blacklist.add(e)

    #remove references to removed art
    for c in [subB]+deps:
        if isinstance(c, Cell):
            c._objects=[e for e in c.objects if e not in blacklist]
            
    #clean heirarchy
    subB.prune()
    
    return (subA, subB)


def relayer(cell, old_layers, new_layer):
    """
    Move any elements in old_layers to new_layer

    :param cell: The :class:`Cell` to process
    :param old_layers: A list of layers to be relayed
    :param new_layer: The layer that ``old_layers`` will be moved to    
    :returns: A copy of the relayered ``Cell``

    TODO: use same labelling scheme found in GdsImport
    """
    new_cell=cell.copy()

    #change layer of art
    for e in new_cell.get_dependencies(True):        
        if not isinstance(e, (Cell, CellReference, CellArray)):
            if e.layer in old_layers:
                e.layer=new_layer

    return new_cell



#===========================================
#functions below here probably no longer work

def chop(polygon, position, axis):
    """
    **BROKEN**: Slice polygon at a given position along a given axis.
    
    Parameters

    polygon : array-like[N][2]
        Coordinates of the vertices of the polygon.
    position : number
        Position to perform the slicing operation along the specified
        axis.
    axis : 0 or 1
        Axis along which the polygon will be sliced.
    
    Returns

    out : tuple[2]
        Each element is a list of polygons (array-like[N][2]).    The first
        list contains the polygons left before the slicing position, and
        the second, the polygons left after that position.
    """
    out_polygons = ([], [])
    polygon = list(polygon)
    while polygon[-1][axis] == position:
        polygon = [polygon[-1]] + polygon[:-1]
    cross = list(numpy.sign(numpy.array(polygon)[:, axis] - position))
    bnd = ([], [])
    i = 0
    while i < len(cross):
        if cross[i - 1] * cross[i] < 0:
            if axis == 0:
                polygon.insert(i, [position, polygon[i - 1][1] + (position - polygon[i - 1][0]) * float(polygon[i][1] - polygon[i - 1][1]) / (polygon[i][0] - polygon[i - 1][0])])
            else:
                polygon.insert(i, [polygon[i - 1][0] + (position - polygon[i - 1][1]) * float(polygon[i][0] - polygon[i - 1][0]) / (polygon[i][1] - polygon[i - 1][1]), position])
            cross.insert(i, 0)
            bnd[1 * (cross[i + 1] > 0)].append(i)
            i += 2
        elif cross[i] == 0:
            j = i + 1
            while cross[j] == 0:
                j += 1
            if cross[i - 1] * cross[j] < 0:
                bnd[1 * (cross[j] > 0)].append(j - 1)
            i = j + 1
        else:
            i += 1
    if len(bnd[0]) == 0:
        out_polygons[1 * (numpy.sum(cross) > 0)].append(polygon)
        return out_polygons
    bnd = (numpy.array(bnd[0]), numpy.array(bnd[1]))
    bnd = (list(bnd[0][numpy.argsort(numpy.array(polygon)[bnd[0], 1 - axis])]),
           list(bnd[1][numpy.argsort(numpy.array(polygon)[bnd[1], 1 - axis])]))
    cross = numpy.ones(len(polygon), dtype=int)
    cross[bnd[0]] = -2
    cross[bnd[1]] = -1
    i = 0
    while i < len(polygon):
        if cross[i] > 0 and polygon[i][axis] != position:
            start = i
            side = 1 * (polygon[i][axis] > position)
            out_polygons[side].append([polygon[i]])
            cross[i] = 0
            nxt = i + 1
            if nxt == len(polygon):
                nxt = 0
            boundary = True
            while nxt != start:
                out_polygons[side][-1].append(polygon[nxt])
                if cross[nxt] > 0:
                    cross[nxt] = 0
                if cross[nxt] < 0 and boundary:
                    j = bnd[cross[nxt] + 2].index(nxt)
                    nxt = bnd[-cross[nxt] - 1][j]
                    boundary = False
                else:
                    nxt += 1
                    if nxt == len(polygon):
                        nxt = 0
                    boundary = True
        i += 1
    return out_polygons


def slice(layer, objects, position, axis, datatype=0):
    """
    **BROKEN** Slice polygons and polygon sets at given positions along an axis.

    Parameters

    layer : integer, list
        The GDSII layer numbers for the elements between each division.  If
        the number of layers in the list is less than the number of divided
        regions, the list is repeated.
    objects : ``Polygon``, ``PolygonSet``, or list
        Operand of the slice operation.  If this is a list, each element
        must be a ``Polygon``, ``PolygonSet``, ``CellReference``,
        ``CellArray``, or an array-like[N][2] of vertices of a polygon.
    position : number or list of numbers
        Positions to perform the slicing operation along the specified
        axis.
    axis : 0 or 1
        Axis along which the polygon will be sliced.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and 255).

    Returns

    out : list[N] of PolygonSet
        Result of the slicing operation, with N = len(positions) + 1.  Each
        PolygonSet comprises all polygons between 2 adjacent slicing
        positions, in crescent order.

    Examples

    >>> ring = gdspy.Round(1, (0, 0), 10, inner_radius = 5)
    >>> result = gdspy.slice(1, ring, [-7, 7], 0)
    >>> cell.add(result[1])
    """
    if (layer.__class__ != [].__class__):
        layer = [layer]
    if (objects.__class__ != [].__class__):
        objects = [objects]
    if (position.__class__ != [].__class__):
        position = [position]
    position.sort()
    result = [[] for i in range(len(position) + 1)]
    polygons = []
    for obj in objects:
        if isinstance(obj, Boundary):
            polygons.append(obj.points)
        elif isinstance(obj, Elements):
            polygons += obj.polygons
        elif isinstance(obj, CellReference) or isinstance(obj, CellArray):
            polygons += obj.get_polygons()
        else:
            polygons.append(obj)
    for i, p in enumerate(position):
        nxt_polygons = []
        for pol in polygons:
            (pol1, pol2) = chop(pol, p, axis)
            result[i] += pol1
            nxt_polygons += pol2
        polygons = nxt_polygons
    result[-1] = polygons
    for i in range(len(result)):
        result[i] = Elements(layer[i % len(layer)], result[i], datatype)
    return result


def boolean(layer, objects, operation, max_points=199, datatype=0, eps=1e-13):
    """
    **BROKEN** Execute any boolean operation on polygons and polygon sets.

    Parameters

    layer : integer
        The GDSII layer number for the resulting element.
    objects : array-like
        Operands of the boolean operation. Each element of this array must
        be a ``Polygon``, ``PolygonSet``, ``CellReference``, ``CellArray``,
        or an array-like[N][2] of vertices of a polygon.
    operation : function
        Function that accepts as input ``len(objects)`` integers.  Each
        integer represents the incidence of the corresponding ``object``.
        The function must return a bool or integer (interpreted as bool).
    max_points : integer
        If greater than 4, fracture the resulting polygons to ensure they
        have at most ``max_points`` vertices. This is not a tessellating
        function, so this number should be as high as possible. For
        example, it should be set to 199 for polygons being drawn in GDSII
        files.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and 255).
    eps : positive number
        Small number to be used as tolerance in intersection and overlap
        calculations.

    Returns

    out : PolygonSet
        Result of the boolean operation.

    Notes

    Since ``operation`` receives a list of integers as input, it can be
    somewhat more general than boolean operations only. See the examples
    below.

    Because of roundoff errors there are a few cases when this function
    can cause segmentation faults. If that happens, increasing the value
    of ``eps`` might help.

    Examples

    >>> circle = gdspy.Round(0, (0, 0), 10)
    >>> triangle = gdspy.Round(0, (0, 0), 12, number_of_points=3)
    >>> bad_poly = gdspy.L1Path(1, (0, 0), '+y', 2,
            [6, 4, 4, 8, 4, 5, 10], [-1, -1, -1, 1, 1, 1])
    >>> union = gdspy.boolean(1, [circle, triangle],
            lambda cir, tri: cir or tri)
    >>> intersection = gdspy.boolean(1, [circle, triangle],
            lambda cir, tri: cir and tri)
    >>> subtraction = gdspy.boolean(1, [circle, triangle],
            lambda cir, tri: cir and not tri)
    >>> multi_xor = gdspy.boolean(1, [badPath], lambda p: p % 2)
    """
    polygons = []      
    indices = [0]
    special_function = False
    for obj in objects:
        if isinstance(obj, ElementBase):
            polygons.append(obj.points)
            indices.append(indices[-1] + 1)
        elif isinstance(obj, Elements):
            special_function = True
            polygons += obj.polygons
            indices.append(indices[-1] + len(obj.polygons))
        elif isinstance(obj, CellReference) or isinstance(obj, CellArray):
            special_function = True
            a = obj.get_polygons()
            polygons += a
            indices.append(indices[-1] + len(a))
        else:
            polygons.append(obj)
            indices.append(indices[-1] + 1)
    if special_function:
        result = boolext.clip(polygons, lambda *p: operation(*[sum(p[indices[ia]:indices[ia + 1]]) for ia in range(len(indices) - 1)]), eps)
    else:
        result = boolext.clip(polygons, operation, eps)
    return None if result is None else Elements(layer, result, datatype, False).fracture(max_points)
